from bespin.amazon.ec2 import get_instances_in_asg_by_lifecycle_state, resume_processes, suspend_processes
from bespin.amazon.s3 import delete_key_from_s3, list_keys_from_s3_path, upload_file_to_s3
from bespin.amazon.sqs import get_all_deployment_messages
from bespin.errors import NoSuchStack, BadDeployment
from bespin.option_spec import stack_specs
from bespin.layers import Layers
from bespin import helpers as hp

from input_algorithms.spec_base import NotSpecified
from input_algorithms.meta import Meta
import logging
import json
import os

log = logging.getLogger("bespin.actions.builder")

class Builder(object):
    def sanity_check(self, stack, stacks, ignore_deps=False, checked=None):
        """Check for missing environment variables in all the stacks"""
        if checked is None:
            checked = []

        if stack.stack_name in checked:
            return

        log.info("Sanity checking %s", stack.key_name)
        stack.find_missing_env()
        stack_specs.stack_json_spec().normalise(Meta({}, []), stack.stack_json_obj)
        if stack.params_json_obj:
            stack_specs.params_json_spec().normalise(Meta({}, []), stack.params_json_obj)

        if not ignore_deps and not stack.ignore_deps:
            for dependency in stack.dependencies(stacks):
                self.sanity_check(stacks[dependency], stacks, ignore_deps, checked + [stack.stack_name])

        if any(stack.build_after):
            for dependency in stack.build_after:
                self.sanity_check(stacks[dependency], stacks, ignore_deps, checked + [stack.stack_name])

    def deploy_stack(self, stack, stacks, credentials, made=None, ignore_deps=False):
        """Make us an stack"""
        self.sanity_check(stack, stacks, ignore_deps=ignore_deps)

        made = made or {}

        if stack.name in made:
            return

        if stack.name not in stacks:
            raise NoSuchStack(looking_for=stack.name, available=stacks.keys())

        if not ignore_deps and not stack.ignore_deps:
            for dependency in stack.dependencies(stacks):
                self.deploy_stack(stacks[dependency], stacks, credentials, made=made, ignore_deps=True)

            for dependency in stack.dependencies(stacks):
                stacks[dependency].cloudformation.wait()

        # Should have all our dependencies now
        log.info("Making stack for '%s' (%s)", stack.name, stack.stack_name)
        self.build_stack(stack)
        made[stack.name] = True

        if stack.sns_confirmation is not NotSpecified and stack.sns_confirmation.straight_after:
            stack.cloudformation.wait()
            self.confirm_deployment(stack, credentials)

        if any(stack.build_after):
            stack.cloudformation.wait()
            for dependency in stack.build_after:
                self.deploy_stack(stacks[dependency], stacks, credentials, made=made, ignore_deps=True)

            for dependency in stack.build_after:
                stacks[dependency].cloudformation.wait()

        stack.cloudformation.wait()
        if stack.sns_confirmation is not NotSpecified and not stack.sns_confirmation.straight_after:
            self.confirm_deployment(stack, credentials)

    def layered(self, stacks, only_pushable=False):
        """Yield layers of stacks"""
        if only_pushable:
            operate_on = dict((stack, instance) for stack, instance in stacks.items() if instance.stack_index)
        else:
            operate_on = stacks

        layers = Layers(operate_on, all_stacks=stacks)
        layers.add_all_to_layers()
        return layers.layered

    def build_stack(self, stack):
        print("Building - {0}".format(stack.stack_name))
        print(json.dumps(stack.params_json_obj, indent=4))
        if stack.skip_update_if_equivalent and all(check.resolve() for check in stack.skip_update_if_equivalent):
            log.info("Stack is determined to be the same, not updating")
        else:
            stack.create_or_update()

    def find_missing_build_env(self, stack):
        for artifact in stack.artifacts.values():
            artifact.find_missing_env()

    def publish_artifacts(self, stack, credentials):
        # Find missing env before doing anything
        self.find_missing_build_env(stack)

        # Iterate over each artifact we need to build
        for key, artifact in stack.artifacts.items():
            # Gather our environment variables
            environment = dict(env.pair for env in artifact.build_env)

            # Create a temporary file to tar to
            with hp.a_temp_file() as temp_tar_file:
                # Make the artifact
                hp.generate_tar_file(temp_tar_file, artifact.commands + artifact.paths + artifact.files
                    , environment=environment
                    , compression=artifact.compression_type
                    )
                log.info("Finished generating artifact: {0}".format(key))

                # Upload the artifact
                upload_file_to_s3(credentials, temp_tar_file.name, artifact.upload_to.format(**environment))

    def confirm_deployment(self, stack, credentials):
        autoscaling_group_id = stack.sns_confirmation.autoscaling_group_id
        asg_physical_id = stack.cloudformation.map_logical_to_physical_resource_id(autoscaling_group_id)
        instances_to_check = get_instances_in_asg_by_lifecycle_state(credentials, asg_physical_id, lifecycle_state="InService")

        environment = dict(env.pair for env in stack.sns_confirmation.env)
        version_message = stack.sns_confirmation.version_message.format(**environment)

        failed = []
        success = []
        attempt = 0

        for _ in hp.until(action="Checking for valid deployment actions"):
            messages = get_all_deployment_messages(credentials, stack.sns_confirmation.deployment_queue)

            # Look for success and failure in the messages
            for message in messages:
                log.info("Message received %s", message['output'])

                # Ignore the messages for instances outside this deployment
                if message['instance_id'] in instances_to_check:
                    if message['output'] == version_message:
                        log.info("Deployed instance %s", message['instance_id'])
                        success.append(message['instance_id'])
                    else:
                        log.info("Failed to deploy instance %s", message['instance_id'])
                        log.info("Failure Message: ", "%s", message['output'])
                        failed.append(message['instance_id'])

            # Stop trying if we have all the instances
            if set(failed + success) == set(instances_to_check):
                break

            # Record the iteration of checking for a valid deployment
            attempt += 1
            log.info("Completed attempt %s of checking for a valid deployment state", attempt)

        if success:
            log.info("Succeeded to deploy %s", success)
        if failed:
            log.info("Failed to deploy %s", failed)
            raise BadDeployment("")

        log.info("All instances have been confirmed to be deployed with version_message [%s]!", version_message)

    def clean_old_artifacts(self, stack, credentials):
        # Find missing env before doing anything
        self.find_missing_build_env(stack)

        # Iterate over each artifact we need to clean
        for key, artifact in stack.artifacts.items():
            environment = dict(env.pair for env in artifact.build_env)

            # Get contents of bucket
            artifact_path = os.path.dirname(artifact.upload_to.format(**environment))
            artifact_keys = list_keys_from_s3_path(credentials, artifact_path)

            # Get all the time stamps and determine the files to delete
            timestamps = list(map(lambda x: x.last_modified, artifact_keys))
            timestamps.sort()
            keys_to_del = timestamps[:-artifact.history_length]

            # Iterate through all the artifacts deleting any ones flagged for deletion
            for artifact_key in artifact_keys:
                if artifact_key.last_modified in keys_to_del:
                    log.info("Deleting artifact %s ", artifact_key.name)
                    delete_key_from_s3(credentials, artifact_key, stack.bespin.dry_run)

    def suspend_cloudformation_actions(self, stack, credentials):
        autoscaling_group_id = stack.sns_confirmation.autoscaling_group_id
        asg_physical_id = stack.cloudformation.map_logical_to_physical_resource_id(autoscaling_group_id)

        suspend_processes(credentials, asg_physical_id)

        log.info("Suspended Processes on AutoScaling Group %s", asg_physical_id)

    def resume_cloudformation_actions(self, stack, credentials):
        autoscaling_group_id = stack.sns_confirmation.autoscaling_group_id
        asg_physical_id = stack.cloudformation.map_logical_to_physical_resource_id(autoscaling_group_id)

        resume_processes(credentials, asg_physical_id)

        log.info("Resumed Processes on AutoScaling Group %s", asg_physical_id)

    def print_artifact_location(self, stack, artifact):
        # Find missing env before doing anything
        self.find_missing_build_env(stack)

        # Iterate over each artifact we need to clean
        for key, artifact_obj in stack.artifacts.items():
            if key == artifact:
                environment = dict(env.pair for env in artifact_obj.build_env)
                print(artifact_obj.upload_to.format(**environment))



import logging

log = logging.getLogger("bespin.amazon.ec2")

def get_instances_in_asg_by_lifecycle_state(credentials, asg_physical_id, lifecycle_state=None):
    instances = []

    asg = credentials.autoscale.get_all_groups(names=[asg_physical_id])
    for instance in asg[0].instances:
        if lifecycle_state is None or lifecycle_state == instance.lifecycle_state:
            instances.append(instance.instance_id)

    return instances

def resume_processes(credentials, asg_physical_id):
    credentials.autoscale.suspend_processes(asg_physical_id, ["ScheduledActions"])

def suspend_processes(credentials, asg_physical_id):
    credentials.autoscale.suspend_processes(asg_physical_id, ["ScheduledActions"])

def display_instances(credentials, asg_physical_id):
    log.info("Finding instances")
    asg = credentials.autoscale.get_all_groups(names=[asg_physical_id])
    instance_ids = [inst.instance_id for inst in asg[0].instances]
    print("Found {0} instances".format(len(instance_ids)))
    print("=" * 20)
    for instance in credentials.ec2.get_only_instances(instance_ids=instance_ids):
        print("{0}\t{1}\t{2}".format(instance.id, instance.private_ip_address, instance.state))



"""
We define a custom spec type here for interpreting list specifications.
"""

from input_algorithms.spec_base import NotSpecified, Spec, formatted
from input_algorithms.errors import BadSpecValue
import six

class many_item_formatted_spec(Spec):
    """
    A spec for something that is many items
    Either a list or a string split by ":"

    If it's a string it will split by ':'
    Otherwise if it's a list, then it will use as is
    and will complain if it has two many values

    It will use determine_<num> on any value that is still NotSpecified after
    splitting the val.

    And will use alter_<num> on all values after they have been formatted.

    Where <num> is 1 indexed index of the value in the spec specifications.

    Finally, create_result is called at the end to create the final result from
    the determined/formatted/altered values.
    """
    specs = []
    value_name = None
    seperators = ":"
    optional_specs = []

    def setup(self, *args, **kwargs):
        if not self.value_name:
            self.value_name = self.__class__.__name__

    def normalise(self, meta, val):
        original_val = val
        if isinstance(val, (list, tuple)):
            vals = val
            dividers = [':']

        elif isinstance(val, six.string_types):
            vals = []
            dividers = []
            while val and any(seperator in val for seperator in self.seperators):
                for seperator in self.seperators:
                    if seperator in val:
                        nxt, val = val.split(seperator, 1)
                        vals.append(nxt)
                        dividers.append(seperator)
                        break
            vals.append(val)

            if not vals:
                vals = [val]
                dividers=[None]

        elif isinstance(val, dict):
            if len(val) > 1:
                raise BadSpecValue("Value as a dict must only be one item", got=val, meta=meta)
            vals = val.items()[0]
            dividers = [':']

        else:
            vals = [val]
            dividers = []

        if len(vals) < len(self.specs) > len(self.specs) + len(self.optional_specs):
            raise BadSpecValue("The value is a list with the wrong number of items"
                , got=val
                , got_length=len(val)
                , min_length=len(self.specs)
                , max_length=len(self.specs) + len(self.optional_specs)
                , looking_at = self.value_name
                )

        formatted_vals = []
        for index, spec in enumerate(self.specs + self.optional_specs):
            index += 1
            expected_type = NotSpecified
            if isinstance(spec, (list, tuple)):
                spec, expected_type = spec

            val = NotSpecified
            if index <= len(vals):
                val = vals[index-1]

            if len(vals) < index:
                vals.append(val)
            val = getattr(self, "determine_{0}".format(index), lambda *args: val)(*list(vals) + [meta, val])
            spec = getattr(self, "spec_wrapper_{0}".format(index), lambda spec, *args: spec)(spec, *list(vals) + [meta, val, dividers])

            if (index-1 < len(self.specs) or val is not NotSpecified) and (expected_type is NotSpecified or not isinstance(val, expected_type)):
                if getattr(self, "formatter", None):
                    val = formatted(spec, formatter=self.formatter).normalise(meta, val)
                else:
                    val = spec.normalise(meta, val)

            func = getattr(self, "alter_{0}".format(index), lambda *args: val)
            altered = func(*(formatted_vals[:index] + [val, meta, original_val]))
            if index <= len(vals):
                vals[index-1] = altered
            formatted_vals.append(altered)

        return self.create_result(*list(formatted_vals) + [meta, original_val, dividers])




from bespin.errors import BadS3Bucket

from six.moves.urllib.parse import urlparse
from contextlib import contextmanager
from collections import namedtuple
from filechunkio import FileChunkIO

import humanize
import logging
import boto
import sys
import os

from boto.s3.key import Key

log = logging.getLogger("bespin.amazon.s3")

S3Location = namedtuple("S3Location", ["bucket", "key", "full"])

def s3_location(value):
    """Return us an s3 location record type"""
    info = urlparse(value)
    if info.scheme != "s3":
        raise ValueError("S3 location must be a valid s3 url\tgot={0}".format(value))

    bucket = info.netloc
    if not bucket:
        raise ValueError("S3 location must be a valid s3 url\tgot={0}".format(value))

    key = info.path
    return S3Location(bucket, key, value)

@contextmanager
def a_multipart_upload(bucket, key):
    mp = None
    try:
        mp = bucket.initiate_multipart_upload(key)
        yield mp
    except:
        if mp:
            mp.cancel_upload()
        raise
    else:
        if mp:
            mp.complete_upload()

def get_bucket(credentials, bucket_name):
    try:
        return credentials.s3.get_bucket(bucket_name)
    except boto.exception.S3ResponseError as error:
        if error.status in (404, 403):
            raise BadS3Bucket("Bucket either doesn't exist or isn't available to you", name=bucket_name)
        else:
            raise

def list_keys_from_s3_path(credentials, query_path):
    query = s3_location(query_path)
    bucket = get_bucket(credentials, query.bucket)
    return bucket.list(prefix="rca-contract")

def delete_key_from_s3(credentials, key, dry_run):
    if dry_run:
        print("Would delete {0}".format(key.name))
    else:
        print("Deleted {0}".format(key.name))
        key.delete()

def upload_file_to_s3(credentials, source_filename, destination_path):
    source_size = os.stat(source_filename).st_size
    if source_size > 30000000:
        upload_file_to_s3_in_parts(credentials, source_filename, destination_path)
    else:
        upload_file_to_s3_as_single(credentials, source_filename, destination_path)

def upload_file_to_s3_in_parts(credentials, source_filename, destination_path):
    source_file = open(source_filename, 'rb')
    destination_file = s3_location(destination_path)

    source = os.path.abspath(source_file.name)
    source_size = os.stat(source).st_size
    log.info("Uploading from %s (%s) to %s in parts", source, humanize.naturalsize(source_size), destination_file.full)

    bucket = get_bucket(credentials, destination_file.bucket)

    try:
        with a_multipart_upload(bucket, destination_file.key) as mp:
            for chunk, offset, length in determine_chunks(source_size, min_chunk=5242881):
                with FileChunkIO(source, 'r', offset=offset, bytes=length) as fp:
                    log.info("Uploading chunk %s (%s)", chunk+1, humanize.naturalsize(length))
                    mp.upload_part_from_file(fp, part_num=chunk+1)

    except boto.exception.S3ResponseError as error:
        if error.status is 403:
            log.error("Seems you are unable to edit this location :(")
            sys.exit(1)
        else:
            raise

    log.info("Finished uploading")

def determine_chunks(total_size, min_chunk=5242881):
    offset = 0
    offsets = []
    while offset < total_size:
        if total_size - offset < min_chunk:
            break
        offsets.append(offset)
        offset += min_chunk
    offsets.append(total_size)
    log.info("Broken up into %s chunks", len(offsets) - 1)

    for index, offset in enumerate(offsets[:-1]):
        nxt = offsets[index+1]
        yield index, offset, nxt - offset

def upload_file_to_s3_as_single(credentials, source_filename, destination_path):
    destination_file = s3_location(destination_path)

    source = os.path.abspath(source_filename)
    source_size = os.stat(source).st_size
    log.info("Uploading from %s (%s) to %s in one go", source, humanize.naturalsize(source_size), destination_file.full)

    bucket = get_bucket(credentials, destination_file.bucket)
    key = Key(bucket)
    key.name = destination_file.key
    key.set_contents_from_filename(source_filename)
    key.close()



from bespin.amazon.cloudformation import Cloudformation
from bespin.helpers import memoized_property
from bespin.errors import BespinError

import boto.sts
import boto.iam
import boto.s3
import boto.sqs
import boto.ec2.autoscale

from input_algorithms.spec_base import NotSpecified
import logging
import boto
import os

log = logging.getLogger("iam_syncr.amazon")

class Credentials(object):
    def __init__(self, region, account_id, assume_role):
        self.region = region
        self.account_id = account_id
        self.assume_role = assume_role

    def verify_creds(self):
        """Make sure our current credentials are for this account and set self.connection"""
        if getattr(self, "_verified", None):
            return

        if self.assume_role is not NotSpecified:
            self.assume()

        log.info("Verifying amazon credentials")
        try:
            connection = boto.iam.connect_to_region(self.region)
        except boto.exception.NoAuthHandlerFound:
            raise BespinError("Export AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY before running this script (your aws credentials)")

        try:
            result = connection.list_roles(max_items=1)
        except boto.exception.BotoServerError as error:
            if error.status == 403:
                raise BespinError("Your credentials aren't allowed to look at iam :(")
            else:
                raise

        roles = result["list_roles_response"]["list_roles_result"]["roles"]
        if not roles:
            raise BespinError("There are no roles in your account, I can't figure out the account id")

        amazon_account_id = roles[0]['arn'].split(":")[4]
        if int(self.account_id) != int(amazon_account_id):
            raise BespinError("Please use credentials for the right account", expect=self.account_id, got=amazon_account_id)

        self._verified = True

    def assume(self):
        log.info("Assuming role as aws:arn:iam::%s:%s", self.account_id, self.assume_role)

        for name in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_SECURITY_TOKEN', 'AWS_SESSION_TOKEN']:
            if name in os.environ and not os.environ[name]:
                del os.environ[name]

        try:
            conn = boto.sts.connect_to_region(self.region)
        except boto.exception.NoAuthHandlerFound:
            raise BespinError("Export AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY before running this script (your aws credentials)")

        creds = conn.assume_role("arn:aws:iam::{0}:{1}".format(self.account_id, self.assume_role), "bespin")
        creds_dict = creds.credentials.to_dict()

        os.environ['AWS_ACCESS_KEY_ID'] = creds_dict["access_key"]
        os.environ['AWS_SECRET_ACCESS_KEY'] = creds_dict["secret_key"]
        os.environ['AWS_SECURITY_TOKEN'] = creds_dict["session_token"]
        os.environ['AWS_SESSION_TOKEN'] = creds_dict["session_token"]

    @memoized_property
    def s3(self):
        self.verify_creds()
        return boto.s3.connect_to_region(self.region)

    @memoized_property
    def ec2(self):
        self.verify_creds()
        return boto.ec2.connect_to_region(self.region)

    @memoized_property
    def autoscale(self):
        self.verify_creds()
        return boto.ec2.autoscale.connect_to_region(self.region)

    @memoized_property
    def sqs(self):
        self.verify_creds()
        return boto.sqs.connect_to_region(self.region)

    @memoized_property
    def iam(self):
        self.verify_creds()
        return boto.iam.connect_to_region(self.region)

    def cloudformation(self, stack_name, region):
        self.verify_creds()
        return Cloudformation(stack_name, region)




from input_algorithms.errors import BadSpec, BadSpecValue
from delfick_error import DelfickError, ProgrammerError

class BespinError(DelfickError): pass

# Explicitly make these errors in this context
BadSpec = BadSpec
BadSpecValue = BadSpecValue
ProgrammerError = ProgrammerError

class BadConfiguration(BespinError):
    desc = "Bad configuration"

class BadOptionFormat(BespinError):
    desc = "Bad option format"

class BadTask(BespinError):
    desc = "Bad task"

class BadOption(BespinError):
    desc = "Bad Option"

class NoSuchKey(BespinError):
    desc = "Couldn't find key"

class NoSuchStack(BespinError):
    desc = "Couldn't find stack"

class BadStack(BespinError):
    desc = "Bad stack"

class BadS3Bucket(BespinError):
    desc = "Bad S3 Bucket"

class BadJson(BespinError):
    desc = "Bad json"

class FailedStack(BespinError):
    desc = "Something about an stack failed"

class BadYaml(BespinError):
    desc = "Invalid yaml file"

class BadResult(BespinError):
    desc = "A bad result"

class BadAmazon(BespinError):
    desc = "Amazon says no"

class StackDoesntExist(BadAmazon):
    desc = "Missing stack"

class UserQuit(BespinError):
    desc = "User quit the program"

class BadDockerConnection(BespinError):
    desc = "Failed to connect to docker"

class BadSQSMessage(BespinError):
    desc= "Failed to decode message"

class StackDepCycle(BespinError):
    desc = "Stack dependency cycle"

class MissingOutput(BespinError):
    desc = "Couldn't find an output"

class MissingFile(BespinError):
    desc = "Couldn't find a file"

class CouldntKill(BespinError):
    desc = "Failed to kill a process"

class BadDirectory(BadSpecValue):
    desc = "Expected a path to a directory"

class BadFilename(BadSpecValue):
    desc = "Expected a path to a filename"

class BadDeployment(BespinError):
    desc = "Failed to get all the correct deployment messages"

class DeprecatedFeature(BadSpecValue):
    desc = "Feature is deprecated"




"""
Custom specifications for the different types of stack options.

The idea is that these understand the conditions around representation of the
options.
"""

from bespin.option_spec.stack_objs import StaticVariable, DynamicVariable, Environment, Skipper
from bespin.option_spec.specs import many_item_formatted_spec
from bespin.option_spec.artifact_objs import ArtifactCommand
from bespin.option_spec.artifact_objs import ArtifactPath
from bespin.formatter import MergedOptionStringFormatter

from input_algorithms.spec_base import NotSpecified
from input_algorithms import spec_base as sb

class var_spec(many_item_formatted_spec):
    value_name = "Variable"
    specs = [sb.string_or_int_as_string_spec()]
    optional_specs = [sb.string_or_int_as_string_spec()]
    formatter = MergedOptionStringFormatter
    seperators = "|"

    def create_result(self, variable, variable_value, meta, val, dividers):
        if variable_value is NotSpecified:
            return StaticVariable(variable)
        else:
            stack = variable
            return DynamicVariable(stack, variable_value)

class artifact_path_spec(many_item_formatted_spec):
    value_name = "Artifact Path"
    specs = [sb.string_spec(), sb.string_spec()]
    formatter = MergedOptionStringFormatter

    def create_result(self, host_path, artifact_path, meta, val, dividers):
        return ArtifactPath(host_path, artifact_path)

class env_spec(many_item_formatted_spec):
    value_name = "Environment Variable"
    seperators = [':', '=']

    specs = [sb.string_spec()]
    optional_specs = [sb.string_or_int_as_string_spec()]
    formatter = MergedOptionStringFormatter

    def create_result(self, env_name, other_val, meta, val, dividers):
        """Set default_val and set_val depending on the seperator"""
        args = [env_name]
        if other_val is NotSpecified:
            other_val = None
        if not dividers:
            args.extend([None, None])
        elif dividers[0] == ':':
            args.extend([other_val, None])
        elif dividers[0] == '=':
            args.extend([None, other_val])
        return Environment(*args)

class skipper_spec(many_item_formatted_spec):
    value_name = "Skip specification"
    spec = lambda: sb.delayed(var_spec())
    specs = [spec(), spec()]

    def create_result(self, var1, var2, meta, val, dividers):
        return Skipper(var1, var2)

formatted_string = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter)

artifact_command_spec = lambda : sb.create_spec(ArtifactCommand
    , copy = sb.listof(artifact_path_spec())
    , modify = sb.dictof(sb.string_spec(), sb.set_options(append=sb.listof(formatted_string)))
    , command = formatted_string
    , add_into_tar = sb.listof(artifact_path_spec())
    )

params_json_spec = lambda: sb.listof(sb.set_options(
      ParameterKey = sb.required(sb.any_spec())
    , ParameterValue = sb.required(sb.any_spec())
    ))

stack_json_spec = lambda: sb.set_options(
      Resources = sb.required(sb.dictof(sb.string_spec(), sb.set_options(Type=sb.required(sb.string_spec()), Properties=sb.optional_spec(sb.dictionary_spec()))))
    , Parameters = sb.optional_spec(sb.dictof(sb.string_spec(), sb.dictionary_spec()))
    , Outputs = sb.optional_spec(sb.dictof(sb.string_spec(), sb.dictionary_spec()))
    )




"""
This is the entry point of Bespin itself.

The overview object is responsible for collecting configuration, knowing default
tasks, and for starting the chosen task.
"""

from bespin.errors import BadConfiguration, BadTask, BadYaml
from bespin.option_spec.bespin_specs import BespinSpec
from bespin.option_spec.task_objs import Task
from bespin.tasks import available_tasks

from input_algorithms.dictobj import dictobj
from option_merge import MergedOptions
from input_algorithms.meta import Meta
from option_merge import Converter
import logging
import yaml
import os

log = logging.getLogger("bespin.executor")

class Overview(object):
    def __init__(self, configuration_file, logging_handler=None):
        self.logging_handler = logging_handler

        self.configuration = self.collect_configuration(configuration_file)
        self.configuration_folder = os.path.dirname(os.path.abspath(configuration_file))
        self.setup_logging_theme()

    def start(self, cli_args, available_tasks=None):
        """Do the bespin stuff"""
        if "stacks" not in self.configuration:
            raise BadConfiguration("Didn't find any stacks in the configuration")
        if not self.configuration.get("environments"):
            raise BadConfiguration("Didn't find any environments configuration")

        bespin = cli_args.pop("bespin")
        self.configuration.update(
            { "$@": bespin.get("extra", "")
            , "bespin": bespin
            , "config_root": self.configuration_folder
            , "environment": bespin.get("environment")
            }
        , source = "<cli>"
        )

        self.configuration.converters.activate()
        bespin = self.configuration["bespin"]
        tasks = self.find_tasks()
        task = bespin["chosen_task"]
        if task not in tasks:
            raise BadTask("Unknown task", task=task, available=tasks.keys())
        stack = getattr(tasks[task], "stack", bespin["chosen_stack"])

        tasks[task].run(self, cli_args, stack, available_tasks=available_tasks)

    ########################
    ###   THEME
    ########################

    def setup_logging_theme(self):
        """
        Setup a logging theme

        Currently there is only ``light`` and ``dark`` which consists of a difference
        in color for INFO level messages.
        """
        if "term_colors" not in self.configuration:
            return

        if not getattr(self, "logging_handler", None):
            log.warning("Told to set term_colors but don't have a logging_handler to change")
            return

        colors = self.configuration.get("term_colors")
        if not colors:
            return

        if colors not in ("light", "dark"):
            log.warning("Told to set colors to a theme we don't have\tgot=%s\thave=[light, dark]", colors)
            return

        # Haven't put much effort into actually working out more than just the message colour
        if colors == "light":
            self.logging_handler._column_color['%(message)s'][logging.INFO] = ('cyan', None, False)
        else:
            self.logging_handler._column_color['%(message)s'][logging.INFO] = ('blue', None, False)

    ########################
    ###   CONFIG
    ########################

    def read_yaml(self, filepath):
        """Read in a yaml file and return as a python object"""
        try:
            if os.stat(filepath).st_size == 0:
                return {}
            return yaml.load(open(filepath))
        except yaml.parser.ParserError as error:
            raise BadYaml("Failed to read yaml", location=filepath, error_type=error.__class__.__name__, error="{0}{1}".format(error.problem, error.problem_mark))

    def get_committime_or_mtime(self, location):
        """Get the commit time of some file or the modified time of of it if can't get from git"""
        return os.path.getmtime(location)

    def home_dir_configuration_location(self):
        """Return the location of the configuration in the user's home directory"""
        return os.path.expanduser("~/.bespin.yml")

    def collect_configuration(self, configuration_file):
        """Return us a MergedOptions with this configuration and any collected configurations"""
        errors = []

        result = self.read_yaml(configuration_file)

        bespin_spec = BespinSpec()
        configuration = MergedOptions(dont_prefix=[dictobj])

        home_dir_configuration = self.home_dir_configuration_location()
        sources = [home_dir_configuration, configuration_file]

        def make_mtime_func(source):
            """Lazily calculate the mtime to avoid wasted computation"""
            return lambda: self.get_committime_or_mtime(source)

        for source in sources:
            if source is None or not os.path.exists(source):
                continue

            try:
                result = self.read_yaml(source)
            except BadYaml as error:
                errors.append(error)
                continue

            if not result:
                continue

            result["mtime"] = make_mtime_func(source)

            if "stacks" in result:
                stacks = result.pop("stacks")
                stacks = dict(
                      (stack, MergedOptions.using(configuration.root(), val, converters=configuration.converters, source=source))
                      for stack, val in stacks.items()
                    )
                result["stacks"] = stacks

            configuration.update(result, dont_prefix=[dictobj], source=source)

            for stack in result.get('stacks', {}).keys():
                self.make_stack_converters(stack, configuration, bespin_spec)

        def convert_bespin(path, val):
            log.info("Converting %s", path)
            meta = Meta(path.configuration, [("bespin", "")])
            configuration.converters.started(path)
            return bespin_spec.bespin_spec.normalise(meta, val)

        bespin_converter = Converter(convert=convert_bespin, convert_path=["bespin"])
        configuration.add_converter(bespin_converter)

        def convert_environments(path, val):
            log.info("Converting %s", path)
            meta = Meta(path.configuration, [("environments", "")])
            configuration.converters.started(path)
            return bespin_spec.environments_spec.normalise(meta, val)

        environments_converter = Converter(convert=convert_environments, convert_path=["environments"])
        configuration.add_converter(environments_converter)

        if errors:
            raise BadConfiguration("Some of the configuration was broken", _errors=errors)

        return configuration

    def make_stack_converters(self, stack, configuration, bespin_spec):
        """Make converters for this stack and add them to the configuration"""
        def convert_stack(path, val):
            log.info("Converting %s", path)
            configuration.converters.started(path)
            environment = configuration['bespin'].environment

            config_as_dict = configuration.as_dict(ignore=["stacks"])
            val_as_dict = val.as_dict(ignore=["stacks"])
            environment_as_dict = configuration[["environments", environment]].as_dict()

            base = path.configuration.root().wrapped()
            everything = path.configuration.root().wrapped()

            base.update(config_as_dict)
            everything.update(config_as_dict)

            base.update(val_as_dict)
            everything[path] = val_as_dict

            base.update(environment_as_dict)
            everything.update(environment_as_dict)

            for thing in (base, everything):
                thing["bespin"] = configuration["bespin"]
                thing["environment"] = environment
                thing["configuration"] = configuration

            meta = Meta(everything, [("stacks", ""), (stack, "")])
            return bespin_spec.stack_spec.normalise(meta, base)

        converter = Converter(convert=convert_stack, convert_path=["stacks", stack])
        configuration.add_converter(converter)

        def convert_tasks(path, val):
            spec = bespin_spec.tasks_spec(available_tasks)
            meta = Meta(path.configuration.root(), [('stacks', ""), (stack, ""), ('tasks', "")])
            configuration.converters.started(path)
            return spec.normalise(meta, val)

        converter = Converter(convert=convert_tasks, convert_path=["stacks", stack, "tasks"])
        configuration.add_converter(converter)

    ########################
    ###   TASKS
    ########################

    def default_tasks(self):
        """Return default tasks"""
        def t(name, description, action=None, **options):
            if not action:
                action = name
            return (name, Task(action, description=description, options=options, label="Bespin"))
        return dict([
              t("show", "Show the available stacks")
            , t("deploy", "Deploy a particular stack")
            , t("bastion", "SSH into the bastion")
            , t("instances", "Find and ssh into the instances")
            , t("list_tasks", "List the available tasks")
            , t("clean_old_artifacts", "Cleans old artifacts from S3")
            , t("confirm_deployment", "Checks all instances were correctly deployed to")
            , t("sanity_check", "Sanity check a stack and it's dependencies")
            , t("print_artifact_location", "Shows where the artifact will be for this environment")
            , t("publish_artifacts", "Makes and uploads the artifacts for a stack to S3")
            , t("resume_cloudformation_actions", "Suspends all schedule actions on a cloudformation stack")
            , t("suspend_cloudformation_actions", "Resumes all schedule actions on a cloudformation stack")
            ])

    def find_tasks(self, configuration=None):
        """Find the custom tasks and record the associated stack with each task"""
        if configuration is None:
            configuration = self.configuration

        tasks = self.default_tasks()
        for stack in list(configuration["stacks"]):
            path = configuration.path(["stacks", stack, "tasks"], joined="stacks.{0}.tasks".format(stack))
            nxt = configuration.get(path, {})
            for task in nxt.values():
                task.specify_stack(stack)
            tasks.update(nxt)

        return tasks



"""
The functionality itself for each task.

Each task is specified with the ``a_task`` decorator and indicates whether it's
necessary to provide the task with the object containing all the stacks and/or
one specific stack object.
"""

from bespin.amazon.ec2 import display_instances
from bespin.actions.builder import Builder
import itertools
import logging

log = logging.getLogger("bespin.tasks")

available_tasks = {}
class a_task(object):
    """Records a task in the ``available_tasks`` dictionary"""
    def __init__(self, needs_artifact=False, needs_stack=False, needs_stacks=False, needs_credentials=False):
        self.needs_artifact = needs_artifact
        self.needs_stack = needs_stack
        self.needs_stacks = needs_stack or needs_stacks
        self.needs_credentials = needs_credentials

    def __call__(self, func):
        available_tasks[func.__name__] = func
        func.needs_artifact = self.needs_artifact
        func.needs_stack = self.needs_stack
        func.needs_stacks = self.needs_stacks
        func.needs_credentials = self.needs_credentials
        return func

@a_task()
def list_tasks(overview, configuration, **kwargs):
    """List the available_tasks"""
    print("Available tasks to choose from are:")
    print("Use the --task option to choose one")
    print("")
    keygetter = lambda item: item[1].label
    tasks = sorted(overview.find_tasks().items(), key=keygetter)
    for label, items in itertools.groupby(tasks, keygetter):
        print("--- {0}".format(label))
        print("----{0}".format("-" * len(label)))
        sorted_tasks = sorted(list(items), key=lambda item: len(item[0]))
        max_length = max(len(name) for name, _ in sorted_tasks)
        for key, task in sorted_tasks:
            print("\t{0}{1} :-: {2}".format(" " * (max_length-len(key)), key, task.description or ""))
        print("")

@a_task(needs_stacks=True)
def show(overview, configuration, stacks, **kwargs):
    """Show what stacks we have"""
    flat = configuration.get("bespin.flat", False)
    only_pushable = configuration.get("bespin.only_pushable", False)

    for index, layer in enumerate(Builder().layered(stacks, only_pushable=only_pushable)):
        if flat:
            for _, stack in layer:
                print(stack.stack_name)
        else:
            print("Layer {0}".format(index))
            for _, stack in layer:
                print("    {0}".format(stack.display_line()))
            print("")

@a_task(needs_stack=True, needs_credentials=True)
def deploy(overview, configuration, stacks, stack, credentials, artifact, **kwargs):
    """Deploy a particular stack"""
    Builder().deploy_stack(stack, stacks, credentials)

@a_task(needs_stack=True, needs_credentials=True)
def publish_artifacts(overview, configuration, stacks, stack, credentials, artifact, **kwargs):
    """Build and publish an artifact"""
    Builder().publish_artifacts(stack, credentials)

@a_task(needs_stack=True, needs_credentials=True)
def clean_old_artifacts(overview, configuration, stacks, stack, credentials, artifact, **kwargs):
    """Cleanup old artifacts"""
    Builder().clean_old_artifacts(stack, credentials)

@a_task(needs_stack=True, needs_credentials=True)
def confirm_deployment(overview, configuration, stacks, stack, credentials, artifact, **kwargs):
    """Confirm deployment via SNS notification for each instance"""
    Builder().confirm_deployment(stack, credentials)

@a_task(needs_stack=True, needs_artifact=True)
def print_artifact_location(overview, configuration, stacks, stack, credentials, artifact, **kwargs):
    """Shows where the artifact will be for this environment"""
    Builder().print_artifact_location(stack, artifact)

@a_task(needs_stack=True, needs_credentials=True)
def suspend_cloudformation_actions(overview, configuration, stacks, stack, credentials, artifact, **kwargs):
    """Suspends all schedule actions on a cloudformation stack"""
    Builder().suspend_cloudformation_actions(stack, credentials)

@a_task(needs_stack=True, needs_credentials=True)
def resume_cloudformation_actions(overview, configuration, stacks, stack, credentials, artifact, **kwargs):
    """Resumes all schedule actions on a cloudformation stack"""
    Builder().resume_cloudformation_actions(stack, credentials)

@a_task(needs_stack=True)
def sanity_check(overview, configuration, stacks, stack, artifact, **kwargs):
    """Sanity check a stack and it's dependencies"""
    Builder().sanity_check(stack, stacks)
    log.info("All the stacks are sane!")

@a_task(needs_stack=True, needs_credentials=True)
def instances(overview, configuration, stacks, stack, artifact, **kwargs):
    """Find and ssh into instances"""
    if artifact is None:
        asg_physical_id = stack.cloudformation.map_logical_to_physical_resource_id(stack.ssh.autoscaling_group_name)
        display_instances(stack.bespin.credentials, asg_physical_id)
    else:
        stack.ssh.ssh_into(artifact, configuration["$@"])

@a_task(needs_stack=True)
def bastion(overview, configuration, stacks, stack, **kwargs):
    """SSH into the bastion"""
    stack.ssh.ssh_into_bastion(configuration["$@"])




from bespin.errors import MissingOutput, BadOption, BadStack, BadJson, BespinError
from bespin.errors import StackDoesntExist
from bespin import helpers as hp

from input_algorithms.spec_base import NotSpecified
from input_algorithms.dictobj import dictobj
import logging
import shlex
import json
import stat
import six
import os

log = logging.getLogger("bespin.option_spec.stack_objs")

class Stack(dictobj):
    fields = [
          "bespin", "name", "key_name", "environment", "stack_json", "params_json"
        , "vars", "stack_name", "env", "build_after", "ignore_deps", "artifacts"
        , "skip_update_if_equivalent", "tags", "sns_confirmation", "ssh"
        ]

    def dependencies(self, stacks):
        for value in self.vars.values():
            if hasattr(value, "stack") and not isinstance(value.stack, six.string_types):
                yield value.stack.key_name

    @property
    def build_after(self):
        for stack in self._build_after:
            if isinstance(stack, six.string_types):
                yield stack
            else:
                yield stack.key_name

    @build_after.setter
    def build_after(self, val):
        self._build_after = val

    def display_line(self):
        return "Stack {0}".format(self.stack_name)

    def find_missing_env(self):
        """Find any missing environment variables"""
        missing = []
        for e in self.env:
            if e.default_val is None and e.set_val is None:
                if e.env_name not in os.environ:
                    missing.append(e.env_name)

        if missing:
            raise BadOption("Some environment variables aren't in the current environment", missing=missing)

    @property
    def cloudformation(self):
        if not hasattr(self, "_cloudformation"):
            self._cloudformation = self.bespin.credentials.cloudformation(self.stack_name, self.bespin.region)
        return self._cloudformation

    @property
    def params_json_obj(self):
        with open(self.params_json) as fle:
            params = fle.read()

        for thing in (self.vars.items(), [env.pair for env in self.env], self.artifact_vars):
            for var, value in thing:
                key = "XXX_{0}_XXX".format(var.upper())
                if key in params:
                    if not isinstance(value, six.string_types):
                        value = value.resolve()
                    params = params.replace(key, value)

        try:
            return json.loads(params)
        except ValueError as error:
            raise BadJson("Couldn't parse the parameters", filename=self.params_json, stack=self.key_name, error=error)

    @hp.memoized_property
    def stack_json_obj(self):
        try:
            return json.load(open(self.stack_json))
        except ValueError as error:
            raise BadJson("Couldn't parse the stack", filename=self.stack_json, stack=self.key_name, error=error)

    @property
    def artifact_vars(self):
        for name, artifact in self.artifacts.items():
            for var, value in artifact.vars:
                value = value.format(**dict(env.pair for env in self.env))
                yield var, value

    def create_or_update(self):
        log.info("Creating or updating the stack (%s)", self.stack_name)
        status = self.cloudformation.wait()

        if not status.exists:
            log.info("No existing stack, making one now")
            self.cloudformation.create(self.stack_json_obj, self.params_json_obj, self.tags.as_dict() or None)
        elif status.complete:
            log.info("Found existing stack, doing an update")
            self.cloudformation.update(self.stack_json_obj, self.params_json_obj)
        else:
            raise BadStack("Stack could not be updated", name=self.stack_name, status=status.name)

class StaticVariable(dictobj):
    fields = ["value"]

    def resolve(self):
        return self.value

class DynamicVariable(dictobj):
    fields = ["stack", "output", ("credentials", None), ("region", None)]

    def resolve(self):
        if isinstance(self.stack, six.string_types):
            outputs = self.credentials.cloudformation(self.stack, self.region).outputs
        else:
            outputs = self.stack.cloudformation.outputs

        if self.output not in outputs:
            raise MissingOutput(wanted=self.output, available=outputs.keys())

        return outputs[self.output]

class Environment(dictobj):
    """A single environment variable, and it's default or set value"""
    fields = ["env_name", ("default_val", None), ("set_val", None)]

    @property
    def pair(self):
        """Get the name and value for this environment variable"""
        if self.set_val is not None:
            return self.env_name, self.set_val
        elif self.default_val is not None:
            return self.env_name, os.environ.get(self.env_name, self.default_val)
        else:
            return self.env_name, os.environ[self.env_name]

class Skipper(dictobj):
    fields = ["var1", "var2"]

    def resolve(self):
        try:
            v1 = self.var1().resolve()
        except StackDoesntExist:
            return False

        try:
            v2 = self.var2().resolve()
        except StackDoesntExist:
            return False

        return v1 and v2 and v1 == v2

class SSH(dictobj):
    fields = [
          "user", "bastion", "bastion_key_location"
        , "instance_key_location", "autoscaling_group_name"
        , "instance_key_path", "bastion_key_path"
        ]

    def ssh_into_bastion(self, extra_args):
        if not os.path.exists(self.bastion_key_path):
            log.error("Didn't find a bastion key, please download the key")
            print("Bastion key can be found at {0}".format(self.bastion_key_location))
            print("Download it to {0}".format(self.bastion_key_path))
            raise BespinError("Couldn't find an ssh key for the bastion")

        os.chmod(self.bastion_key_path, 0)
        os.chmod(self.bastion_key_path, stat.S_IRUSR)

        command = "ssh {0}@{1} -i {2} -o IdentitiesOnly=true".format(self.user, self.bastion, self.bastion_key_path)
        parts = shlex.split(command)
        os.execvp(parts[0], parts)

    def ssh_into(self, ip_address, extra_args):
        proxy = ""
        error = False
        if self.bastion is not NotSpecified:
            log.info("Logging into %s via %s", ip_address, self.bastion)
            if not os.path.exists(self.bastion_key_path):
                log.error("Didn't find a bastion key, please download the key")
                print("Bastion key can be found at {0}".format(self.bastion_key_location))
                print("Download it to {0}".format(self.bastion_key_path))
                error = True
            proxy = '-o ProxyCommand="ssh {0}@{1} -W %h:%p -i {2} -o IdentitiesOnly=true"'.format(self.user, self.bastion, self.bastion_key_path)
        else:
            log.info("Logging into %s", ip_address)

        if not os.path.exists(self.instance_key_path):
            log.error("Didn't find a instance key, please download the key")
            print("Instance key can be found at {0}".format(self.instance_key_location))
            print("Download it to {0}".format(self.instance_key_path))
            error = True

        if error:
            raise BespinError("Couldn't find ssh keys")

        os.chmod(self.instance_key_path, 0)
        os.chmod(self.instance_key_path, stat.S_IRUSR)
        os.chmod(self.bastion_key_path, 0)
        os.chmod(self.bastion_key_path, stat.S_IRUSR)

        command = "ssh -o ForwardAgent=false -o IdentitiesOnly=true {0} -i {1} {2}@{3} {4}".format(proxy, self.instance_key_path, self.user, ip_address, extra_args)
        parts = shlex.split(command)
        os.execvp(parts[0], parts)




"""
Some helper functions for running subprocesses and ensuring they don't hang or
stick around.
"""
from bespin.errors import CouldntKill

import subprocess
import logging
import signal
import shlex
import fcntl
import time
import os

log = logging.getLogger("bespin.processes")

def read_non_blocking(stream):
    """Read from a non-blocking stream"""
    if stream:
        while True:
            nxt = ''
            try:
                nxt = stream.readline()
            except IOError:
                pass

            if nxt:
                yield nxt
            else:
                break

def command_output(command, *command_extras, **kwargs):
    """Get the output from a command"""
    output = []
    cwd = kwargs.get("cwd", None)
    args = shlex.split(' '.join([command] + list(command_extras)))
    timeout = kwargs.get("timeout", 10)
    verbose = kwargs.get("verbose", False)

    log_level = log.info if verbose else log.debug
    log_level("Running command\targs=%s", args)
    process = subprocess.Popen(args, stderr=subprocess.STDOUT, stdout=subprocess.PIPE, cwd=cwd)

    fl = fcntl.fcntl(process.stdout, fcntl.F_GETFL)
    fcntl.fcntl(process.stdout, fcntl.F_SETFL, fl | os.O_NONBLOCK)

    start = time.time()
    while True:
        if time.time() - start > timeout:
            break
        if process.poll() is not None:
            break
        for nxt in read_non_blocking(process.stdout):
            nxt_out = nxt.decode("utf8").strip()
            output.append(nxt_out)
            if verbose:
                print(nxt_out)

        time.sleep(0.01)

    attempted_sigkill = False
    if process.poll() is None:
        start = time.time()
        log.error("Command taking longer than timeout (%s). Terminating now\tcommand=%s", timeout, args)
        process.terminate()

        while True:
            if time.time() - start > timeout:
                break
            if process.poll() is not None:
                break
            for nxt in read_non_blocking(process.stdout):
                output.append(nxt.decode("utf8").strip())
            time.sleep(0.01)

        if process.poll() is None:
            log.error("Command took another 5 seconds after terminate, so sigkilling it now")
            os.kill(process.pid, signal.SIGKILL)
            attempted_sigkill = True

    for nxt in read_non_blocking(process.stdout):
        output.append(nxt.decode("utf8").strip())

    if process.poll() is not 0 and attempted_sigkill:
        raise CouldntKill("Failed to sigkill hanging process", pid=process.pid, command=args, output="\n".join(output))

    if process.poll() != 0:
        log.error("Failed to run command\tcommand=%s", args)

    return output, process.poll()




from bespin.errors import BadSQSMessage

import logging
import json
import time

log = logging.getLogger("bespin.amazon.sqs")

# Get all the messages of the queue, dropping those that are not deployment messages
def get_all_deployment_messages(credentials, sqs_url, timeout=60, sleep=2):
    start = time.time()
    messages = []

    q = credentials.sqs.get_queue(sqs_url)
    while q.count() > 0 and not (time.time() - start > timeout):
        raw_messages = credentials.sqs.receive_message(q, number_messages=1)

        if len(raw_messages) > 0:
            raw_message = raw_messages[0]
            encoded_message = json.loads(raw_message.get_body())['Message']

            #q.delete_message(raw_message)

            message = decode_message(encoded_message)

            if message is not None:
                messages.append(message)

        time.sleep(sleep)

    return messages

# Takes a message that is : separated and maps it to a Dict
def decode_message(encoded_message):
    # Check if this is a valid message
    if encoded_message.count(':') < 2:
        raise BadSQSMessage("Less than two colons", msg=encoded_message)

    result, instance_id, output = encoded_message.split(':', 2)
    return {
        'result': result,
        'instance_id': instance_id,
        'output': output
    }



"""
We have here the object representing a task.

Tasks contain a reference to the functionality it provides (in ``bespin.tasks``)
as well as options that are used to override those in the stack it's attached to.
"""

from bespin.amazon.credentials import Credentials
from bespin.errors import BadOption

from input_algorithms.spec_base import NotSpecified
from input_algorithms.dictobj import dictobj
from option_merge import MergedOptions

class Task(dictobj):
    """
    Used to add extra options associated with the task and to start the action
    from ``bespin.tasks``.

    Also responsible for complaining if the specified action doesn't exist.
    """
    fields = [("action", "run"), ("label", "Project"), ("options", None), ("overrides", None), ("description", "")]

    def run(self, overview, cli_args, stack, available_tasks=None):
        """Run this task"""
        if available_tasks is None:
            from bespin.tasks import available_tasks
        task_func = available_tasks[self.action]
        configuration = MergedOptions.using(overview.configuration, dont_prefix=overview.configuration.dont_prefix, converters=overview.configuration.converters)

        if self.options:
            if stack:
                configuration.update({"stacks": {stack: self.options}})
            else:
                configuration.update(self.options)

        configuration.update(cli_args, source="<cli>")

        if self.overrides:
            overrides = {}
            for key, val in self.overrides.items():
                overrides[key] = val
                if isinstance(val, MergedOptions):
                    overrides[key] = dict(val.items())
            overview.configuration.update(overrides)

        stacks = None
        if task_func.needs_stacks:
            environment = configuration["bespin"].environment
            if not environment:
                raise BadOption("Please specify an environment")
            if configuration["environments"].get(environment) is None:
                raise BadOption("No configuration found for specified environment", environment=environment)

            stacks = self.determine_stack(stack, overview, configuration, needs_stack=task_func.needs_stack)
            if stack:
                stack = stacks[stack]

        credentials = None
        if task_func.needs_credentials:
            assume_role = NotSpecified if configuration["bespin"].no_assume_role else configuration["bespin"].assume_role
            credentials = Credentials(
                  configuration["bespin"].region
                , configuration["environments"][environment].account_id
                , assume_role
                )
            configuration["bespin"].credentials = credentials

        artifact = configuration["bespin"].chosen_artifact or None
        if task_func.needs_artifact and not artifact:
                raise BadOption("Please specify an artifact")

        return task_func(overview, configuration, stacks=stacks, stack=stack, credentials=credentials, artifact=artifact)

    def determine_stack(self, stack, overview, configuration, needs_stack=True):
        """Complain if we don't have an stack"""
        stacks = configuration["stacks"]

        available = None
        available = stacks.keys()

        if needs_stack:
            if not stack:
                info = {}
                if available:
                    info["available"] = list(available)
                raise BadOption("Please use --stack to specify a stack to use", **info)

            if stack not in stacks:
                raise BadOption("No such stack", wanted=stack, available=list(stacks.keys()))

        return stacks

    def specify_stack(self, stack):
        """Specify the stack this task belongs to"""
        self.stack = stack




from setuptools import setup, find_packages
from bespin import VERSION

setup(
      name = "bespin"
    , version = VERSION
    , packages = ['bespin'] + ['bespin.%s' % pkg for pkg in find_packages('bespin')]
    , include_package_data = True

    , install_requires =
      [ "delfick_error==1.6.1"
      , "option_merge==0.9.1"
      , "input_algorithms==0.4.3"
      , "argparse==1.3.0"

      , "six"
      , "humanize"

      , "boto==2.34.0"
      , "pyYaml==3.10"
      , "rainbow_logging_handler==2.2.2"
      , "FileChunkIO==1.6"
      ]

    , extras_require =
      { "tests":
        [ "noseOfYeti>=1.5.0"
        , "nose"
        , "mock"
        ]
      }

    , entry_points =
      { 'console_scripts' :
        [ 'bespin = bespin.executor:main'
        ]
      }

    # metadata for upload to PyPI
    , url = "https://github.com/realestate-com-au/bespin"
    , author = "Stephen Moore"
    , author_email = "stephen.moore@rea-group.com"
    , description = "Opinionated wrapper around boto that reads yaml"
    , license = "MIT"
    , keywords = "cloudformation boto"
    )




"""
Here we define the yaml specification for Besin options, task options and stack
options.

The specifications are responsible for sanitation, validation and normalisation.
"""

from input_algorithms.spec_base import (
      formatted, defaulted, any_spec, dictionary_spec, dictof, listof, required
    , string_spec, overridden, boolean, file_spec, optional_spec, integer_spec, or_spec
    , valid_string_spec, create_spec, string_choice_spec, filename_spec as orig_filename_spec
    )

from bespin.option_spec import task_objs, stack_objs, stack_specs, artifact_objs
from bespin.formatter import MergedOptionStringFormatter
from bespin.helpers import memoized_property

from input_algorithms.dictobj import dictobj
from input_algorithms import validators

import six

class filename_spec(orig_filename_spec):
    def setup(self, spec):
        self.spec = spec

    def normalise_either(self, meta, val):
        val = self.spec.normalise(meta, val)
        return super(filename_spec, self).normalise_filled(meta, val)

class Bespin(dictobj):
    fields = [
          "dry_run", "assume_role", "flat", "config", "chosen_stack", "no_assume_role"
        , "region", "environment", "chosen_artifact", "chosen_task", "extra", "interactive"
        ]

class SNSConfirmation(dictobj):
    fields = ["version_message", "env", "straight_after", "deployment_queue", "autoscaling_group_id"]

class Environment(dictobj):
    fields = ["account_id", "vars"]

class other_options(dictobj):
    fields = ["run", "create", "build"]

class BespinSpec(object):
    """Knows about bespin specific configuration"""

    @memoized_property
    def task_name_spec(self):
        """Just needs to be ascii"""
        return valid_string_spec(
              validators.no_whitespace()
            , validators.regexed("^[a-zA-Z][a-zA-Z0-9-_\.]*$")
            )

    def tasks_spec(self, available_actions, default_action="run"):
        """Tasks for a particular stack"""
        return dictof(
              self.task_name_spec
            , create_spec(task_objs.Task
                , action = defaulted(string_choice_spec(available_actions, "No such task"), default_action)
                , options = dictionary_spec()
                , overrides = dictionary_spec()
                , description = string_spec()
                )
            )

    @memoized_property
    def environments_spec(self):
        """Spec for each environment options"""
        return dictof(
              string_spec()
            , create_spec(Environment
                , account_id = required(or_spec(string_spec(), valid_string_spec(validators.regexed("\d+"))))
                , vars = dictionary_spec()
                )
            )

    @memoized_property
    def stack_spec(self):
        """Spec for each stack"""
        return create_spec(stack_objs.Stack
            , bespin = any_spec()

            , name = formatted(defaulted(string_spec(), "{_key_name_1}"), formatter=MergedOptionStringFormatter)
            , key_name = formatted(overridden("{_key_name_1}"), formatter=MergedOptionStringFormatter)
            , stack_name = formatted(defaulted(string_spec(), "{_key_name_1}"), formatter=MergedOptionStringFormatter)
            , environment = formatted(overridden("{environment}"), formatter=MergedOptionStringFormatter)

            , env = listof(stack_specs.env_spec(), expect=stack_objs.Environment)

            , tags = dictionary_spec()

            , stack_json = filename_spec(formatted(defaulted(string_spec(), "{config_root}/{_key_name_1}.json"), formatter=MergedOptionStringFormatter))
            , params_json = filename_spec(formatted(defaulted(string_spec(), "{config_root}/{environment}/{_key_name_1}-params.json"), formatter=MergedOptionStringFormatter))

            , build_after = listof(formatted(string_spec(), formatter=MergedOptionStringFormatter))
            , ignore_deps = defaulted(boolean(), False)

            , vars = dictof(string_spec(), stack_specs.var_spec())

            , skip_update_if_equivalent = listof(stack_specs.skipper_spec())

            , sns_confirmation = optional_spec(create_spec(SNSConfirmation
                , env = listof(stack_specs.env_spec(), expect=stack_objs.Environment)
                , version_message = required(formatted(string_spec(), formatter=MergedOptionStringFormatter))
                , autoscaling_group_id = required(formatted(string_spec(), formatter=MergedOptionStringFormatter))
                , deployment_queue = required(formatted(string_spec(), formatter=MergedOptionStringFormatter))
                , straight_after = defaulted(formatted(boolean(), MergedOptionStringFormatter, expected_type=bool), True)
                ))

            , artifacts = dictof(string_spec(), create_spec(artifact_objs.Artifact
                , compression_type = string_choice_spec(["gz", "xz"])
                , history_length = integer_spec()
                , location_var_name = string_spec()
                , upload_to = formatted(string_spec(), formatter=MergedOptionStringFormatter)
                , commands = listof(stack_specs.artifact_command_spec(), expect=artifact_objs.ArtifactCommand)
                , paths = listof(stack_specs.artifact_path_spec(), expect=artifact_objs.ArtifactPath)
                , files = listof(create_spec(artifact_objs.ArtifactFile
                    , content = formatted(string_spec(), formatter=MergedOptionStringFormatter)
                    , path = formatted(string_spec(), formatter=MergedOptionStringFormatter)
                    ))
                , build_env = listof(stack_specs.env_spec(), expect=stack_objs.Environment)
                ))

            , ssh = optional_spec(create_spec(stack_objs.SSH
                , user = optional_spec(formatted(string_spec(), formatter=MergedOptionStringFormatter))
                , bastion = optional_spec(formatted(string_spec(), formatter=MergedOptionStringFormatter))
                , bastion_key_location = optional_spec(formatted(string_spec(), formatter=MergedOptionStringFormatter))
                , instance_key_location = optional_spec(formatted(string_spec(), formatter=MergedOptionStringFormatter))
                , autoscaling_group_name = required(formatted(string_spec(), formatter=MergedOptionStringFormatter))

                , bastion_key_path = formatted(defaulted(string_spec(), "{config_root}/{environment}/bastion_ssh_key.pem"), formatter=MergedOptionStringFormatter)
                , instance_key_path = formatted(defaulted(string_spec(), "{config_root}/{environment}/ssh_key.pem"), formatter=MergedOptionStringFormatter)
                ))
            )

    @memoized_property
    def bespin_spec(self):
        """Spec for bespin options"""
        formatted_string = formatted(string_spec(), MergedOptionStringFormatter, expected_type=six.string_types)
        formatted_boolean = formatted(boolean(), MergedOptionStringFormatter, expected_type=bool)

        return create_spec(Bespin
            , config = file_spec()

            , assume_role = optional_spec(string_spec())

            , dry_run = defaulted(boolean(), False)
            , flat = defaulted(boolean(), False)
            , environment = optional_spec(string_spec())

            , extra = defaulted(formatted_string, "")
            , region = defaulted(string_spec(), "ap-southeast-2")
            , no_assume_role = defaulted(formatted_boolean, False)

            , chosen_task = defaulted(formatted_string, "list_tasks")
            , chosen_stack = defaulted(formatted_string, "")
            , chosen_artifact = defaulted(formatted_string, "")

            , interactive = defaulted(formatted_boolean, True)
            )

